/*
The package platform is a semester project by Galley, Michel,
that Silaghi, Marius-Calin adviced in Spring 2000.
Several modifications were done subsequently by Silaghi, Marius-Calin.
Copyright (c) 2000, 2001, 2002, 2004
*/
/**
 * @(#)Agent.java  1.0 00/02/04
 *
 * Intelligent Agents Collaborating in Distributed Problem Solving
 *
 * Supervisors:   Marius-Calin Silaghi, Djamila Sam-Haroud
 * Author:        Michel Galley
 */

package platform;

import java.net.*;
import sjavap.*;
import utils.*;
import csp.*;
import java.awt.*;
import java.awt.geom.*;

/**
 * This class encapsulates the agent's main data and processing. Please
 * refer to the "Intelligent Agents Collaborating in Distributed Problem
 * Solving" paper for details on the object model, functional model and
 * dynamic model of this active object.
 *
 * @author Michel Galley
 * @see GenericAgent
 * @see Broker
 * @version 1.0 00/02/04
 */

public class Agent extends GenericAgent{
    // sending notify messages
    public static boolean notifying=true;
    private AgentID brokerID;
    private AgentGUI agentGUI;
    /** The message received in the getMessage method */
    private Message receivedMessage;

    /** Agent's states */
    private static final int INITIAL_PHASE = 1; // state before subscription
    private static final int TRANSITION_PHASE = 2; // state between 2 comput.
    private static final int COMPUTATION_PHASE = 3;
    /** Initial state */
    private int state = INITIAL_PHASE;

    /**
     * Initializes a newly created Agent object.
     *
     * @param  agentName Agent's name
     * @param  knownProtocols Array of classes containing the protocols known
     * by the agent
     * @param  broker_host Broker's IP address
     * @param  broker_port Broker's port number
     * @param  withGUI Indicates if the agent must create a GUI
     */
    public Agent(String agentName,
		 Class[] knownProtocols,
		 InetAddress broker_host, int broker_port, boolean withGUI){
	super(knownProtocols);

	// Launch the server
	server = new Server(this, null, inQueue, protocolInQueue);//, 2001);
	new Thread(server).start();

	// Create all the AgentID objects
        brokerID = new AgentID("Broker", true, broker_host, broker_port);
	agentID = new AgentID(agentName, false,
			      server.getHost(), server.getPort());

	// Launch the GUI
	if(withGUI)
	    agentGUI = new AgentGUI(this);
    }

    /**
     * Initializes a newly created Broker object. No GUI is created.
     *
     * @param  agentName Agent's name
     * @param  knownProtocols Class array of the protocols known by the agent
     * @param  broker_host Broker's IP address
     * @param  broker_port Broker's port number
     */
    public Agent(String agentName,
		 Class[] knownProtocols,
		 InetAddress broker_host, int broker_port){
        this(agentName, knownProtocols, broker_host, broker_port, false);
    }

    /** The run method of the agent. */
    synchronized public void run(){
	select.thread=Thread.currentThread();
	//Kernel.dp1=select;
	int reply;

	//System.out.println("Agent started on host: "+server.getHost()
	//		   +"port: "+server.getPort());

	for(;;) {

	    // All the calls accepted in any situation
	    select.accept(0, _getMessage, 1);
	    select.accept(1, _setProblem, 2);
	    select.accept(2, _getProblem);
	    select.accept(3, _subscribe, 3);
	    select.accept(4, _unsubscribe, 3);

	    switch(state) {
	    // All the calls only accepted during the computation phase
	    case COMPUTATION_PHASE:
		select.accept(5, _sendCommand);
		select.accept(6, _createColorMatrix);
		select.accept(7, _setColor);
            }
	    select.accept(8, _notifyMessage);

	    //System.out.println("Platform-w:");
	    //System.err.println("Platform:"+agentID+" State="+state);
	    reply = Synch_Wait();
	    //System.err.println("Platform:"+agentID+" : Reply="+reply+" State="+state);

	    switch(reply) {

	    // getMessage
	    case 0:
		// According to the new message, the state of the Agent
		// may change. All required processing related to incoming
		// messages are made here.

		if(receivedMessage instanceof Message_StartProtocol) {

		    MySystem.err.println("Treating:"+receivedMessage);
		    // Kill the old protocol if it is still running
		    killProtocol();

		    // Empty the output queue
		    select.call(0, outQueue, outQueue._clear);
		    Synch_Wait();
		    outQueue.clear();

		    // Update counters
		    computationNum = receivedMessage.getComputationNum();
		    messageNum = 1;

		    // Create a new instance of the protocol
		    usedProtocol = ((Message_StartProtocol)receivedMessage)
			.getProtocol();
		    try {
			instanceOfUsedProtocol
			    = (Protocol)usedProtocol.newInstance();
			Class agentClass=
			    instanceOfUsedProtocol.getAgentProtocolClass();
			if(agentClass==null){
			    MySystem.out.println("Unknown agent protocol");
			    MySystem.exit(1);
			}
                        usedProtocol=agentClass;
			instanceOfUsedProtocol
			    = (Protocol)agentClass.newInstance();
			instanceOfUsedProtocol.startParameter=((Message_StartProtocol)receivedMessage).parameters;
		    } catch(Exception e) {
			System.err.println("Cannot create a new instance "+
					   "of the protocol");
			System.exit(-1);
		    }
		    // Start the protocol
		    instanceOfUsedProtocol.setParameters(this, problem,
							protocolInQueue, null);
		    threadProtocol = new Thread(instanceOfUsedProtocol);
		    //threadProtocol.setPriority(Thread.MAX_PRIORITY);
		    threadProtocol.start();

		    // Change state
		    state = COMPUTATION_PHASE;
		}
		else if(receivedMessage instanceof Message_EndProtocol) {
		    // Kill the protocol
		    killProtocol();
		    setComputationNum(getComputationNum()+1);
		    setMessageNum(1);
		    //clear incoming queue
		    server.clearIncomingQueue(this,getComputationNum(),protocolInQueue);
		    MySystem.err.println("Send new subscription");
		    Message_Subscription subscription
			= new Message_Subscription(knownProtocols, brokerID);

		    putOnOutputQueue(subscription);
		    // Change state
		    state = TRANSITION_PHASE;
		}
		else if(receivedMessage instanceof Message_Load) {
		    MySystem.err.println("Treating:"+receivedMessage);
		    problem = (Problem)(((Message_Load)receivedMessage).getArguments());
		    Message_Subscription subscription
		    = new Message_Subscription(knownProtocols, brokerID);
		    putOnOutputQueue(subscription);
		}
		else if(receivedMessage instanceof Message_BrokerShutsDown) {
		    // The broker shuts down;
		    // so does the agent
		    shutDown();
		}
		else {
		    // The message has not been recognized :
		    // put it at the end of the queue
		    MySystem.err.println("Not recognized message:"+receivedMessage);
		    MySystem.exit(1);
		    //inQueue.addMessage(receivedMessage);
		}
		break;

	    // subscribe
	    case 3:
		// Send a subscription to the broker
		if(state == INITIAL_PHASE) {
		    computationNum++; //The broker also increments
		    Message_Subscription subscription
			= new Message_Subscription(knownProtocols, brokerID);

		    putOnOutputQueue(subscription);
		    state = TRANSITION_PHASE;
		}
		break;

	    // sendCommand
	    case 5:
		//MySystem.err.println("Platform puts command on output queue");
		putOnOutputQueue(sentMessage);
		break;

	    // InitColorMatrix
	    case 6:
                //MySystem.out.println("Agent:acc:InitColor:"+sentMessage);
		putOnOutputQueue(sentMessage);
		break;

	    // setColor
	    case 7:
                //MySystem.out.println("Agent:acc:setColor:"+sentMessage);
		putOnOutputQueue(sentMessage);
		break;

	    // notifyMessage
	    case 8:
 		putOnOutputQueue(sentMessage);
		break;

	    // unsubscribe
	    case 4:
		// Send an unsubscription to the broker and shut down
		if(state != INITIAL_PHASE) {
		    Message_Unsubscription unsubscription
			= new Message_Unsubscription(brokerID);

		    putOnOutputQueue(unsubscription);
		}
		shutDown();
		break;
	    }
	}
    }

    /****************
     * Public methods
     ****************/


    /**
     * Return the ID of the broker.
     *
     * @return The ID of the broker (AgentID object).
     */
    public AgentID getBrokerID() { return brokerID; }


    /** sJava identification for the getMessage method. */
    public int _getMessage = 1;
    /**
     * Process the message given to the agent.
     *
     * @param  newMessage A new message
     */
    synchronized public void getMessage(Message newMessage){
	receivedMessage = newMessage;
	Synch_Reply();
    }

    /** sJava identification for the subscribe method. */
    public int _subscribe = 22;
    /**
     * Register the agent to the broker. The agent will be considered
     * as a participant in the computations. The agent submits all the
     * protocols it knows (the protocols are passed to the constructor).
     */
    synchronized public void subscribe() { Synch_Reply(); }


    /** sJava identification for the unsubscribe method. */
    public int _unsubscribe = 23;
    /**
     * Unregister the agent to the broker. The agent won't receive
     * messages anymore.
     */
    synchronized public void unsubscribe() { Synch_Reply(); }

    /** sJava identification for the createColorMatrix method. */
    public int _createColorMatrix = 24;
    /**
     * Send a request to the broker to create a matrix representing the
     * constraint shared by two variables (this matrix is shown in the GUI).
     * Call of this method is a prerequisite to any call of a setColor
     * involving this matrix.
     *
     * @param  var1 Name of the first variable
     * @param  var2 Name of the second variable
     * @param  sizeX Width of the matrix
     * @param  sizeY Height of the matrix
     * @see Agent#setColor
     */
    synchronized public void createColorMatrix(int id, String var1, String var2,
					       int sizeX, int sizeY)
    {
	sentMessage = new Message_CreateColorMatrix(id, var1, var2,
						 sizeX, sizeY, brokerID);
	Synch_Reply();
    }

    synchronized public void createColorMatrix(String var1, String var2,
					       int sizeX, int sizeY)
    {
	sentMessage = new Message_CreateColorMatrix(0, var1, var2,
						 sizeX, sizeY, brokerID);
	Synch_Reply();
    }

    /** sJava identification for the setColor method. */
    public int _setColor = 25;
    /**
     * Put a new color in the cell positioned at (posX, posY) in the
     * matrix (of the GUI) involving both var1 and var2 variables.
     *
     * @param  var1 Name of the first variable
     * @param  var2 Name of the second variable
     * @param  sizeX Width of the matrix
     * @param  sizeY Height of the matrix
     * @param  color Cell's new color
     * @param  oldColor Cell's old color
     * @see Agent#createColorMatrix
     */
    synchronized public void setColor(String var1, String var2,
				      int posX, int posY,
				      Color color, Color oldColor)
    {
	 sentMessage = new Message_SetColor(var1, var2, posX, posY,
					 color, oldColor, brokerID);
	Synch_Reply();
    }
    /**
     * Put a new color in the cells positioned at (posX, posY,
     * posX2,posY2) in the
     * matrix (of the GUI) involving both var1 and var2 variables.
     *
     * @param  var1 Name of the first variable
     * @param  var2 Name of the second variable
     * @param  sizeX Width of the matrix
     * @param  sizeY Height of the matrix
     * @param  color Cell's new color
     * @param  oldColor Cell's old color
     * @see Agent#createColorMatrix
     */
    synchronized public void setColor(String var1, String var2,
				      int posX, int posY,
                                      int posX2, int posY2,
				      Color color, Color oldColor)
    {
	 sentMessage = new Message_SetColor(var1, var2, posX, posY,
                                          posX2, posY2,
					 color, oldColor, brokerID);
	Synch_Reply();
    }

    /**
     * Put a new color in the cells positioned at (posX, posY,
     * posX2,posY2) in the
     * matrix (of the GUI) involving both var1 and var2 variables.
     *
     * @param  var1 Name of the first variable
     * @param  var2 Name of the second variable
     * @param  sizeX Width of the matrix
     * @param  sizeY Height of the matrix
     * @param  color Cell's new color
     * @param  oldColor Cell's old color
     * @see Agent#createColorMatrix
     */
    synchronized public void setColor(String var1, String var2,
				      int table[][],
				      int posX, int posY,
                                      int posX2, int posY2,
				      int destX, int destY,
				      Color color, Color oldColor,
				      Color color0, Color oldColor0){
	 sentMessage =
	     new Message_SetColor(var1, var2,
				  posX, posY,
				  posX2, posY2,
				  table,
				  destX, destY,
				  color, oldColor,
				  color0, oldColor0,
				  brokerID);
	Synch_Reply();
    }
    synchronized public void setColor(String var1,
				      String var2,
				      Range[] size,
				      Range[] oldsize,
				      Color color,
				      Color color0){
	 sentMessage =
	     new Message_SetColor(var1, var2,
				  size, oldsize,
				  color, color0,
				  brokerID);
	Synch_Reply();
    }

    int _notifyMessage = 26;
    /**
     * This method is used to notify the broker that the agent has received
     * a message from another agent.
     */
    synchronized void notifyMessage(Message_NotifyMessage message){
	sentMessage = message;
	Synch_Reply();
    }
}

