/*
The package platform is a semester project by Galley, Michel,
that Silaghi, Marius-Calin adviced in Spring 2000.
Several modifications were done subsequently by Silaghi, Marius-Calin.
Copyright (c) 2000, 2001, 2002, 2004
*/
/*
 * @(#)Broker.java  1.0 00/02/04
 *
 * Intelligent Agents Collaborating in Distributed Problem Solving
 *
 * Supervisors:   Marius-Calin Silaghi, Djamila Sam-Haroud
 * Author:        Michel Galley
 */

package platform;

import csp.*;
import java.net.*;
import java.util.*;
import java.io.*;
import utils.*;

/**
 * This class encapsulates the broker's main data and processing. Please
 * refer to the "Intelligent Agents Collaborating in Distributed Problem
 * Solving" paper for details about the object model, functional model and
 * dynamic model of this active object.
 *
 * @author Michel Galley
 * @see GenericAgent
 * @see Agent
 * @version 1.0 00/02/04
 */

public class Broker extends GenericAgent{
    // List of all subscribed agents
    private AgentList agentList;

    // array of protocols for each participant
    private Hashtable protocolsHashTable;
    // protocols known by all participants
    private Vector usableProtocols;
    /** The message list */
    MessageList messageList = null;

    private BrokerOwner owner;

    private BrokerGUI brokerGUI = null;
    //private MessageList messageList = null;

    private static final int TRANSITION_PHASE = 1;
    private static final int COMPUTATION_PHASE = 2;

    private int state = TRANSITION_PHASE;
    private int maxNbOfAgents = Integer.MAX_VALUE;
    public static final int _getAgentIDs = 11;
    public static final int _setMaxNbOfAgents = 12;
    public static final int _getProtocols = 13;
    public static final int _chooseProtocol = 14;
    public static final int _startProtocol = 15;
    public static final int _endProtocol = 16;
    public static final int _brokerShutsDown = 17;
    public static final int _protocolFinished = 18;
    public static final int _removeSubscriptions = 19;
    public static final int _sendMessage = 20;

    /**
     * Initializes a newly created Broker object.
     *
     * @param  knownProtocols Array of classes containing the protocols
     * known by the broker
     * @param  withGUI Indicates if the broker must create a GUI
     */
    public Broker(Class[] knownProtocols, boolean withGUI) {
	super(knownProtocols);
	agentList = new AgentList();

	if(withGUI) {
	    messageList = new MessageList();
	    Message.initMessageLogging();
	}
       	// Launches the server
	server = new Server(this, messageList, inQueue, protocolInQueue);//, 2000);
	new Thread(server).start();

	agentID = new AgentID("Broker", true,
			      server.getHost(), server.getPort());
	protocolsHashTable = new Hashtable();
	protocolsHashTable.put(this.agentID, this.knownProtocols);
	usableProtocols = new Vector(0);

	if(withGUI) {
	    brokerGUI = new BrokerGUI(this, server.getHost(),
				      server.getPort(), messageList);
	}
	updateUsableProtocols();
    }

    /**
     * Initializes a newly created Broker object. No GUI is created.
     *
     * @param  knownProtocols Class array of the protocols known by the broker
     */
    public Broker(Class[] knownProtocols) { this(knownProtocols, false); }

    /** The run method of the broker. */
    synchronized public void run(){
	select.thread=Thread.currentThread();
	int reply;

	//System.out.println("Broker is running on " + server.getHost()
	//		   +" "+ server.getPort());

	for(;;) {
	    // The calls accepted in any situation
	    select.accept(0, _getMessage, 1);
	    select.accept(1, _setProblem, 2);
	    select.accept(2, _getProblem);
	    select.accept(3, _getAgentIDs);
	    select.accept(4, _brokerShutsDown);
	    select.accept(5, _startProtocol, 3);
	    select.accept(6, _endProtocol, 1);
	    select.accept(7, _chooseProtocol, 2);
	    select.accept(8, _protocolFinished, 4);
	    select.accept(12, _sendMessage);
	    select.accept(13, _removeSubscriptions);

	    switch(state) {
	    // The calls accepted between two computations
	    case TRANSITION_PHASE:
		select.accept(9, _setMaxNbOfAgents);
		select.accept(10, _getProtocols);
		break;
	    // The calls accepted during a computation
	    case COMPUTATION_PHASE:
		select.accept(11, _sendCommand);
	    }

	    reply = Synch_Wait();
	    //System.out.println("Broker : Reply="+reply+" State="+state);

	    switch(reply) {

	    // startProtocol
	    case 5:
		startNewProtocol();
		break;

	    // sendCommand
	    case 11:
		putOnOutputQueue(sentMessage);
		break;

	    // sendMessage
	    case 12:
		//MySystem.err.println("Broker asked to "+sentMessage);
		putOnOutputQueue(sentMessage);
		break;

	    // endProtocol
	    case 6:
		Message_EndProtocol epMessage;

		if(state == COMPUTATION_PHASE) {
		    // Sends the "end protocol" message to each agent
		    for(int index=0; index<agentList.nbOfAgents(); index++) {
			epMessage = new Message_EndProtocol
			    (agentList.getAgentID(index));
			putOnOutputQueue(epMessage);
		    }
		}

		// kill the protocol
		if(brokerGUI!=null)
		    brokerGUI.updateProtocolStatus
			("Trying to kill old protocol : "
			 +instanceOfUsedProtocol);
		killProtocol();
		if(brokerGUI!=null)
		    brokerGUI.updateProtocolStatus
			("No active protocol");
		state = TRANSITION_PHASE;
		break;

	    // protocolFinished
	    case 8:
		Message_EndProtocol pfMessage;

		// The protocol has finished his job :
		// send a "end protocol" message to each agent
		if(state == COMPUTATION_PHASE) {
                    int index;
		    // Sends the "end computation" message to each agent
		    for(index=0; index<agentList.nbOfAgents(); index++) {
			pfMessage = new Message_EndProtocol
			    (agentList.getAgentID(index));
			putOnOutputQueue(pfMessage);
		    }
                    for(index=agentList.nbOfAgents()-1; index>=0; index--) {
                      protocolsHashTable.remove(agentList.getAgentID(index));
                      agentList.removeAgent(agentList.getAgentID(index));
                    }
       		}
		if(brokerGUI!=null)
		    brokerGUI.updateProtocolStatus("No active protocol");
		state = TRANSITION_PHASE;
		break;

	    // brokerShutsDown
	    case 4:
		Message_BrokerShutsDown sdMessage;

		// Sends the "broker shuts down" message to each agent
		for(int index=0; index<agentList.nbOfAgents(); index++) {
		    sdMessage = new Message_BrokerShutsDown
			(agentList.getAgentID(index));
		    putOnOutputQueue(sdMessage);
		}

		// clean up code here if needed

		shutDown();
	    }

	    if((maxNbOfAgents < agentList.nbOfAgents()) &&
	       (usedProtocol != null)) {
		startNewProtocol(); // automatically starts the protocol
	    }
	}
    }

    /****************
     * Public methods
     ****************/

    /**
     * The genericAgent sends a message to the genericAgent whose ID
     * is receiverID. The message format is :
     * <command> [<arg1> ... <argn>]
     *
     * @param  command Command serializable object
     * @param  arguments Arguments serializable object
     * @param  receiverID Destination of the object
     */
    synchronized public void sendMessage(Serializable command,
					 Serializable arguments,
					 AgentID receiverID) {
	sentMessage = new Message_Load(command, arguments, receiverID);
	Synch_Reply();
    }

    /** sJava identification for the getMessage method. */
    /**
     * Process the message given to the broker.
     *
     * @param  newMessage A new message
     */
    synchronized public void getMessage(Message newMessage)
    {
	if(newMessage instanceof Message_Subscription) {
	    // A new agent subscription:
	    AgentID newAgentID = newMessage.getSenderID();

	    // Add new agent to the broker's list
	    agentList.addAgent(newAgentID);
	    if(brokerGUI!=null)
		brokerGUI.updateAgents(agentList.getAgentIDs());
	    protocolsHashTable.put(newAgentID,
			((Message_Subscription)newMessage).getProtocols());
            //if(protocolsHashTable.get(newAgentID)==null){MySystem.exit(1);}
	    updateUsableProtocols();
	}
	else if(newMessage instanceof Message_Unsubscription) {
	    // An agent has unsubscribed:
	    AgentID departedAgentID = newMessage.getSenderID();

	    // Remove the agent from the broker's list
	    agentList.removeAgent(departedAgentID);
	    if(brokerGUI!=null)
		brokerGUI.updateAgents(agentList.getAgentIDs());
	    protocolsHashTable.remove(departedAgentID);
	    updateUsableProtocols();
	}
	Synch_Reply();
    }

    synchronized public void removeSubscriptions(){
	agentList.removeAll();
	if(brokerGUI!=null)
	    brokerGUI.updateAgents(agentList.getAgentIDs());
	Object myProts=(Class[])protocolsHashTable.get(agentID);
	protocolsHashTable.clear();
	protocolsHashTable.put(agentID,myProts);
	updateUsableProtocols();
	Synch_Reply();
    }

    /** sJava identification for the getAgentIDs method. */
    /**
     * Return an array containing all agent IDs.
     *
     * @return An array containing all agent IDs
     */
    synchronized public AgentID[] getAgentIDs()
    {
	AgentID[] ids;

	ids = agentList.getAgentIDs();
	Synch_Reply();
	return ids;
    }

    /** sJava identification for the setMaxNbOfAgents method. */
    /**
     * Sets the maxium number of subscribed agents. If this function is
     * not called, we assume that this number is Integer.MAX_VALUE.
     * If a protocol has been chosen, this protocol is started as soon
     * as the maximum number of subscribed agents is reached.
     * This method cannot be called during a computation.
     *
     * @param  nbOfAgents Maximum number of agents
     */
    synchronized public void setMaxNbOfAgents(int nbOfAgents)
    {
	this.maxNbOfAgents = nbOfAgents;
	Synch_Reply();
    }

    /** sJava identification for the getProtocols method. */
    /**
     * Return an array containing the protocols known by all participants.
     */
    synchronized public Class[] getProtocols() {
	Class[] protocols = (Class[])usableProtocols.toArray(new Class[0]);
	Synch_Reply();
	return protocols;
    }

    /** sJava identification for the chooseProtocol method. */
    /**
     * Let the caller choose a protocol among the protocols
     * returned by getProtocols. If the protocol is not known by all
     * participants, chooseProtocol returns false. Otherwise, it
     * rerturns true.
     *
     * @param  protocol The chosen protocol
     */
    synchronized public boolean chooseProtocol(Class protocol)
    {
	boolean found = false;

	if(state == TRANSITION_PHASE) {
	    for(int index = 0; index < usableProtocols.size(); index++)
		if(protocol == (Class)usableProtocols.elementAt(index)) {
		    this.usedProtocol = protocol;
		    found = true;
		    break;
		}
	    if(found) {
		try {
		    MySystem.out.println("Chosen protocol: "+
					 usedProtocol.newInstance());
		}catch(Exception e) {
		    MySystem.out.println("Unknown protocol:"+usedProtocol);
		}
	    }
	}
	Synch_Reply();
	return found;
    }


    /** sJava identification for the startProtocol method. */
    /**
     * The protocol is started. The broker sends a message to all subscribers
     * asking them to start their protocols (the same class).
     */
    synchronized public void startProtocol(BrokerOwner _owner) {
	owner = _owner;
	Synch_Reply();
    }

    /** sJava identification for the endProtocol method. */
    /**
     * The protocol is stopped. The broker sends a message to all subscribers
     * asking them to stop their protocols.
     */
    synchronized public void endProtocol() { Synch_Reply(); }

    /** sJava identification for the brokerShutsDown method. */
    /**
     * The broker shuts down. The broker sends a message to all
     * subscribers telling them the broker is going to shut down.
     */
    synchronized public void brokerShutsDown() { Synch_Reply(); }

    /** sJava identification for the protocolFinished method. */
    /**
     * Only a protocol can call this method. It informs the broker that
     * the computation has ended.
     */
    synchronized public void protocolFinished(Object result) {
	if(owner!=null)
	    owner.answer(result);
	Synch_Reply();
    }

    /*******************
     * Package method
     *******************/

    /**
     * Put a message on the out queue.
     *
     * @param  message The message to put on the out queue.
     */
    protected void putOnOutputQueue(Message message){
	super.putOnOutputQueue(message);
	if(messageList != null)
	    messageList.addMessage(message);
    }

    /*****************
     * Private methods
     *****************/

    /** Start the protocol */
    private void startNewProtocol(){
	if(state == TRANSITION_PHASE) {

	    // If no protocol selected, choose a default protocol
	    if(usedProtocol == null && usableProtocols.size() > 0)
		// sets the default protocol
		usedProtocol = (Class)usableProtocols.elementAt(0);

	    // If a protocol is selected, launch it
	    if(usedProtocol != null) {

		Message_StartProtocol spMessage;

		// Update counters
		this.computationNum++;
		this.messageNum = 1;

		synchronized(server.queueClear){
		    // Empty the protocol message queue
		    server.clearIncomingQueue(this,computationNum,protocolInQueue);

		    // Empty the input queue
		    //server.clearIncomingQueue(this,computationNum,inQueue);
		}

		// Empty the output queue<-LOADS are there
		//select.call(0, outQueue, outQueue._clear);
		//Synch_Wait();
		//outQueue.clear();


		// Create a new instance of the protocol
		try {
		    instanceOfUsedProtocol
			= (Protocol)usedProtocol.newInstance();
		} catch(Exception e) {
		    System.err.println("Cannot create a new instance "+
				       "of the protocol");
		    System.exit(1);
		}
		//MySystem.out.println("here1");
		try{
		    Class brokerClass=
			instanceOfUsedProtocol.getBrokerProtocolClass();
		    //MySystem.out.println("here bC");
		    if(brokerClass==null){
			MySystem.out.println("Unknown broker protocol");
			MySystem.exit(1);
		    }

		    instanceOfUsedProtocol
			= (Protocol)brokerClass.newInstance();
		}catch(Exception e) {
		    MySystem.out.println("Unknown broker protocol");
		    MySystem.exit(1);
		}

		// Sends the "start protocol" message to each agent

		for(int index=0; index<agentList.nbOfAgents(); index++) {
		    spMessage = new Message_StartProtocol
			(this.usedProtocol, agentList.getAgentID(index));
		    putOnOutputQueue(spMessage);
		}


		// Starts the protocol
		instanceOfUsedProtocol.setParameters(this,
			       	       problem,
				       protocolInQueue,
				       agentList.getAgentIDs());
		threadProtocol = new Thread(instanceOfUsedProtocol);
		//threadProtocol.setPriority(Thread.MAX_PRIORITY);
		threadProtocol.start();
		if(brokerGUI!=null)
		    brokerGUI.updateProtocolStatus
			("Active protocol : "+instanceOfUsedProtocol);
		state = COMPUTATION_PHASE;
	    }
	}
    }

    /** Add to candidates all protocols found in protocolArray. */
    private void addCandidates(Vector candidates, Class[] protocolArray){
	boolean found;

	// used by updateUsableProtocols:
	for(int index1 = 0; index1 < protocolArray.length; index1++) {
	    found = false;
	    for(int index2 = 0; index2 < candidates.size(); index2++)
		if(protocolArray[index1]
		   == (Class)candidates.elementAt(index2))
		    found = true;
	    if(!found)
		candidates.add(protocolArray[index1]);
	}
    }

    /** Remove from candidates all protocols not known by all participants. */
    private void purgeCandidates(Vector candidates, Class[] protocolArray)
    {
	// used by updateUsableProtocols:
	boolean found;
	Class curCandidate;

	for(int cand = candidates.size()-1; cand >= 0; cand--) {
	    curCandidate = (Class)candidates.elementAt(cand);
	    found = false;
	    for(int index = 0; index < protocolArray.length; index++)
		if(protocolArray[index] == curCandidate) {
		    found = true;
		    break;
		}
	    if(!found)
		candidates.remove(cand);
	}
    }

    /** Update the list of protocols known by all participants. */
    private void updateUsableProtocols()
    {
	Vector candidates = new Vector(0); // all protocols
	//AgentID[] agentIDs = agentList.getAgentIDs();
	AgentID[] agentIDs = agentList.getAgentIDs();

	// Create the list of candidate protocols
	for(int index = 0; index < agentIDs.length; index++)
	    addCandidates(candidates, (Class[])protocolsHashTable.get
			  (agentIDs[index]));
	addCandidates(candidates, (Class[])protocolsHashTable.get(agentID));

	// Remove from the list the protocols not known by each agent
	for(int index = 0; index < agentIDs.length; index++)
	    purgeCandidates(candidates, (Class[])protocolsHashTable.get
			    (agentIDs[index]));
	purgeCandidates(candidates, (Class[])protocolsHashTable.get(agentID));

	usableProtocols = candidates;
	if(brokerGUI!=null)
	    brokerGUI.updateProtocols(usableProtocols);
    }
}


