/*
The package platform is a semester project by Galley, Michel,
that Silaghi, Marius-Calin adviced in Spring 2000.
Several modifications were done subsequently by Silaghi, Marius-Calin.
Copyright (c) 2000, 2001, 2002, 2004
*/
/*
 * @(#)ChildMessageList.java  1.0 00/02/04
 *
 * Intelligent Agents Collaborating in Distributed Problem Solving
 *
 * Supervisors:   Marius-Calin Silaghi, Djamila Sam-Haroud
 * Author:        Michel Galley
 */
package platform;

import java.io.*;
import java.util.*;
import utils.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.event.*;

/**
 * This class is used as a ListModel implementation for the JList components
 * containing messages specific to a single agent. It is used jointly with the
 * MessageList class. It accesses the MessageList's message buffer to select
 * messages involving the considered agent. One can only access a part of them
 * through the ListModel interface.
 * <p>This class also holds a collection of MatrixModel objects. These matrices
 * represent information sent by the agents about their problems. When new
 * messages about these matrices are processed by the a ChildMessageList
 * object (createColorMatrix and setColor messages), the corresponding
 * MatrixModel is notified of the change.
 *
 * @author Michel Galley
 * @see MessageList
 * @see BrokerChildGUI
 * @see MatrixModel
 * @version 1.0 00/02/04
 */

class ChildMessageList implements ListModel{
    static final int VISIBLE_SIZE = 5;

    AgentID agentID;
    /** The JList componenents listening to this model. */
    Vector listDataListeners;
    /** Collections of constraints */
    ConstraintListModel constraintList;
    /** Collections of variables */
    //VariablesListModel[] varListModels;
    MessageList parentModel;
    /** The MatrixModel objects */
    Vector matrixModels;

    /** A table of indexes on the parentModel's messageBuffer */
    int[] messageSubList;
    int messageSubListSize;
    int selectedMessage;

    /**
     * Initializes a newly created ChildMessageList object.
     *
     * @param  parentModel The MessageList object
     * @param  agentID ID of the agent observed by this model
     */
    ChildMessageList(MessageList parentModel,
			AgentID agentID){
	this.agentID = agentID;
	this.parentModel = parentModel;
	//varListModels = new VariablesListModel[2];
	//varListModels[0] = new VariablesListModel();
	//varListModels[1] = new VariablesListModel();
	constraintList = new ConstraintListModel();
	matrixModels = new Vector(0);
	listDataListeners = new Vector(0);
	messageSubList = new int[MessageList.BUFFER_SIZE];
	messageSubListSize = 0;
	selectedMessage = 0;
    }

    /**
     * Return the MatrixModel object involving the two variables.
     *
     * @param  var1 The first variable
     * @param  var2 The second variable
     * @return The MatrixModel involving these variables
     */
    MatrixModel getMatrixModel(String var1, String var2){
	for(int i=0; i<matrixModels.size(); i++)
	    if(((((MatrixModel)matrixModels.elementAt(i)).getVariable1()
		 .equals(var1) &&
		 ((MatrixModel)matrixModels.elementAt(i)).getVariable2()
		 .equals(var2))) ||
	       ((((MatrixModel)matrixModels.elementAt(i)).getVariable1()
		 .equals(var2) &&
		 ((MatrixModel)matrixModels.elementAt(i)).getVariable2()
		 .equals(var1))))
		return (MatrixModel)matrixModels.elementAt(i);
	return null;
    }

    /**
     * Return the MatrixModel object involving the two variables
     * (notation: variable1 + "x" + variable2)
     *
     * @param  var12 The string containing the two variables.
     * @return The MatrixModel involving these variables
     */
    MatrixModel getMatrixModel(String var12){
	if(var12 == null) return null;

        int fst = var12.indexOf(")");
	int index = var12.indexOf("x");
	String var1 = var12.substring(fst+2, index),
	    var2 = var12.substring(index+1);
	return getMatrixModel(var1, var2);
    }

    /**
     * Return one of the two VarListModel objects
     *
     * @param  i Index of the VarListModel object to return (0 or 1)
     * @return The VarListModel object
     */
    ConstraintListModel getConstraintListModel() {
      return constraintList;
    }
    //VariablesListModel getVarListModel(int i) { return varListModels[i]; }

    /**
     * Process the new message. If this message is a createColorMatrix
     * message, create a new MatrixModel object. The new variables encountered
     * are put into the varListModels objects.
     *
     * @param  message The new message
     */
    void processMessage(Message message){
	MatrixModel model;
	String[] vars = new String[2];
        int id=0;

	if(message instanceof Message_CreateColorMatrix) {
	    vars[0] = ((Message_CreateColorMatrix)message).getVariable1();
	    vars[1] = ((Message_CreateColorMatrix)message).getVariable2();
            id = ((Message_CreateColorMatrix)message).id;
	    model = getMatrixModel(vars[0], vars[1]);
	    if(model == null) {
	        matrixModels.add(new MatrixModel
				 ((Message_CreateColorMatrix)message));
		//no two constraints with the same variables in an agent
		constraintList.addElement("("+id+") "+vars[0]+"x"+vars[1]);
	    }
	}
	//else if(message instanceof Message_SetColor) {
	//    vars[0] = ((Message_SetColor)message).getVariable1();
	//    vars[1] = ((Message_SetColor)message).getVariable2();
            //MySystem.out.println(message);
	//}


	//varListModels[0].addElement(vars[0]);
	//varListModels[0].addElement(vars[1]);
	//varListModels[1].addElement(vars[0]);
	//varListModels[1].addElement(vars[1]);
    }

    /**
     * Move the message selection. Notify the parentModel that the
     * selected message has changed. In its turn, it will also
     * notify all the other ChildMessageList objects.
     *
     * @param  offset Number of messages to skip (signed value)
     */
    boolean moveSelection(int offset)
    {
	if(selectedMessage+offset < 0 ||
	   selectedMessage+offset >= messageSubListSize)
	    return false;

	parentModel.moveSelection(messageSubList[selectedMessage+offset]
				  -parentModel.selectedMessage);
	return true;
    }

    /**
     * Update the messageSubList buffer from parent's model
     * messageBuffer. This method is generally called when the messageBuffer
     * has changed.
     */
    void updateFromBuffer(){
	messageSubListSize = 0;
	for(int i=0; i<=parentModel.bufferEnd-parentModel.bufferStart; i++){
            AgentID aID=parentModel.messageBuffer[i].getSenderID();
            //if(aID==null) aID=((Message_NotifyMessage)parentModel.messageBuffer[i]).getOrigSenderID();
	    if(aID.equals(agentID))
            //||parentModel.messageBuffer[i].getReceiverID().equals(agentID))
		messageSubList[messageSubListSize++] = i;
        }
	updateListView();
    }

    /** Parent's model messageBuffer has been cleared. */
    void clear()
    {
	updateFromBuffer();
	selectedMessage = 0;

	// Clear all matrix models
	for(int i=0; i<matrixModels.size(); i++)
	    ((MatrixModel)matrixModels.elementAt(i)).clear();
    }

    // The buffer has changed, so that the selectedMessage hos no meaning
    void cancelSelection(){
	selectedMessage = -1;
    }

    /**
     * Refresh the message area after a change has occured in the
     * buffer or in the selection. If the selection has changed,
     * find all the setColor messages between the old selection and the
     * new one, then send all the color changes to the right MatrixModel
     * object.
     */
    void updateListView()
    {
	int oldSelectedMessage = selectedMessage;
	MatrixModel model;
	Message_SetColor message;
	String[] vars = new String[2];


	// Find the right message
	for(int i=messageSubListSize-1; i>=0; i--)
	    if(messageSubList[i] <= parentModel.selectedMessage) {
		selectedMessage = i;
		break;
	    }

	// If the selection has moved
	if(oldSelectedMessage != selectedMessage) {
	    boolean forward = (oldSelectedMessage < selectedMessage);
	    int stop = (forward)?selectedMessage+1:selectedMessage;
	    for(int i=(forward)?oldSelectedMessage+1:oldSelectedMessage;
		i!=stop; i=(forward)?i+1:i-1)

		// Process all the setColor messages and send any
		// changes to the right MatrixModel object
		
		if(parentModel.messageBuffer[messageSubList[i]]
		   instanceof Message_SetColor) {
		    message = (Message_SetColor)parentModel
			.messageBuffer[messageSubList[i]];
		    model = getMatrixModel(message.getVariable1(),
					   message.getVariable2());
		    if(model != null) {
			if(forward) model.setColor(message);
			else model.unsetColor(message);
		    }
		}
	}

	for(int i=0; i<listDataListeners.size(); i++) {
	    ListDataListener listDataListener =
		(ListDataListener)listDataListeners.elementAt(i);
	    listDataListener.contentsChanged
		(new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED,
				   0, VISIBLE_SIZE-1));
	}
    }

   /**************************************************************
     * ListModel implementation :
     * a JList listens to the history changes:
     **************************************************************/

    /**
     * Add a listener to the list that's notified each time a change to the
     * data model occurs. In the current implementation, there can be only
     * one listener.
     *
     * @param  listener The listener to add
     * @see ListData
     */
    public void addListDataListener(ListDataListener listener)
    { listDataListeners.add(listener); }
    /**
     * Remove the listener from the list.
     *
     * @param  listener The listener to remove
     * @see ListData
     */
    public void removeListDataListener(ListDataListener listener)
    { listDataListeners.remove(listener); }
    /**
     * Get the size of the visible part of the message buffer. It is
     * a constant value.
     */
    public int getSize() { return VISIBLE_SIZE; }
    /**
     * Get an element from the visible part of the buffer.
     *
     * @param  index Index of the element to access
     */
    public Object getElementAt(int index)
    {
	int indexTable = selectedMessage-VISIBLE_SIZE/2+index;
	if(indexTable < 0 || indexTable >= messageSubListSize)
	    return parentModel.EMPTY_LINE;
	return parentModel.messageBuffer[messageSubList[indexTable]];
    }
}
