/*
The package platform is a semester project by Galley, Michel,
that Silaghi, Marius-Calin adviced in Spring 2000.
Several modifications were done subsequently by Silaghi, Marius-Calin.
Copyright (c) 2000, 2001, 2002, 2004
*/
/*
 * @(#)GenericAgent.java  1.0 00/02/04
 *
 * Intelligent Agents Collaborating in Distributed Problem Solving
 *
 * Supervisors:   Marius-Calin Silaghi, Djamila Sam-Haroud
 * Author:        Michel Galley
 */

package platform;

import csp.*;
import sjavap.*;
import utils.*;
import java.io.*;
import java.lang.*;


/**
 * This is an abstract class encapsulating all data and processing common
 * to both the broker and the agents. Please refer to the "Intelligent
 * Agents Collaborating in Distributed Problem Solving" paper for details
 * about the object model, functional model and dynamic model of this active
 * object.
 *
 * @author Michel Galley
 * @see GenericAgent
 * @see Agent
 * @version 1.0 00/02/04
 */

public abstract class GenericAgent implements Runnable, MessageReceiver{
    /** GenericAgent's server: it receives messages from other agents. */
    public Server server;
    /** The object that sends messages to other agents. */
    MessageSender messageSender;
    /** The message queue between the Server object and this agent. */
    MessageQueue inQueue;
    /** The message queue between the Server object and the protocol. */
    MessageQueue protocolInQueue;
    /** The message queue between this agent and the MessageSender object. */
    public MessageQueue outQueue;
    /** The sJava kernel. */
    Kernel kernel;

    Message sentMessage;

    Class[] knownProtocols;
    Class usedProtocol = null;
    Protocol instanceOfUsedProtocol = null;
    Thread threadProtocol = null;
    Problem problem = null;
    int outputQueueSize = 0;

    AgentID agentID;
    int computationNum = 0;
    int messageNum = 1;

    public static int MAX_MESSAGES = 100;
    public static int FlowConnectionThreshold = 200;

    public static final int _sendCommand = 2;
    public static final int _setProblem = 3;
    public static final int _getProblem = 4;

 /**    active object support (needs implements Reply)               **/
 /*********************************************************************/
 /* Written by Claude Petitpierre,   LTI-DI-EPFL,           7 Apr 1999*/
 /*********************************************************************/
 /* Modif. by Marius-Calin Silaghi, LIA-DI-EPFL,silaghi@lia.di.epfl.ch*/
 /*  Bug that may lead to deadlock in Synch_Wait(),fixed,  30 Nov 1999*/
 /*********************************************************************/
 /**/ Synch_List select = new Synch_List(this) ;                      //
 /**/ synchronized public void Synch_Reply()                          //
 /**/   { notify(); }                                                 //
 /**/ int now() {return Kernel.kern.currentTime;}                     //
 /**/ int Synch_Wait() {                                              //
 /**/   Kernel.kern.Synch_Request(select);                            //
 /**/   do                                                            //
 /**/     {try {wait();} catch(InterruptedException e){}}             //
 /**/   while(select.nbEntries!=0);                                   //
 /**/   return select.Value; }                                        //
 /*********************** Need not be modified ************************/


    /**
     * Initializes a newly created GenericAgent object.
     *
     * @param  knownProtocols Class array of the protocols known by the agent
     */
    GenericAgent(Class[] knownProtocols){
	/** Launch the sJava kernel */
	kernel = new Kernel();
	this.knownProtocols = knownProtocols;

	/** Launch the messageSender, inQueue, outQueue and protocolInQueue
	    threads. */
	messageSender = new MessageSender(this);
	new Thread(messageSender).start();
	inQueue = new MessageQueue(1, this,"inQueue");
	new Thread(inQueue).start();
	outQueue = new MessageQueue(1, messageSender,"outQueue");
	new Thread(outQueue).start();
	protocolInQueue = new MessageQueue(1, null,"protocolInQueue");
	new Thread(protocolInQueue).start();
    }

    /****************
     * Public methods
     ****************/

    /**
     * Return the ID of the GenericAgent object.
     */
    public AgentID getID() { return agentID; }

    /** sJava identification for the sendCommand method. */
    /**
     * The genericAgent sends a message to the genericAgent whose ID
     * is receiverID. The message format is :
     * <command> [<arg1> ... <argn>]
     *
     * @param  command Command serializable object
     * @param  arguments Arguments serializable object
     * @param  receiverID Destination of the object
     */
    synchronized public void sendCommand(Serializable command,
					 Serializable arguments,
					 AgentID receiverID){
	//MySystem.err.println("Platform receives command");
	sentMessage = new Message_Command(command, arguments, receiverID);
	Synch_Reply();
    }

    /** sJava identification for the setProblem method. */
    /**
     * Set the problem the genericAgent has to solve.
     *
     * @param  file The file containing the problem
     */
    synchronized public boolean setProblem(File file){
        if (!file.exists()) {
	    try {
		System.err.println("loadFile:Inexisting file:"
				   +file.getCanonicalPath());
	    }catch(Exception e){
		System.err.println("loadFile:Inexisting file error");
	    }
	    Synch_Reply();
            return false;
	}

	problem = new Problem();
        if (!problem.loadFile(file))  {
	    System.err.println("loadFile:Error loading file");
	    Synch_Reply();
            return false;
	}

	Synch_Reply();
        return true;
    }

    /** sJava identification for the getProblem method. */
    /**
     * Return The problem the genericAgent has to solve
     *
     * @return The problem the genericAgent has to solve
     */
    synchronized public Problem getProblem(){
	Synch_Reply();
	return problem;
    }

    /*****************
     * Package methods
     *****************/

    /** Flow control */
    boolean overflow(){
	if(!protocolInQueue.hasActivators())
	    return false;
	if((FlowConnectionThreshold>0)&&
	   (protocolInQueue.size()>FlowConnectionThreshold))
	    return true;
	return false;
	/*
	int k=inQueue.size()+
	    protocolInQueue.size();
	int outSize=protocolInQueue.size();
	if(k > 0)
	    if(k +
	       outSize > MAX_MESSAGES){
		return true;
	    }
	if(k==0)
	    if(outSize > MAX_MESSAGES){
		//I can block if sending to a smaller
		String crtTarget=messageSender.getReceiverName();
		if((crtTarget!=null)&&
		   (crtTarget.compareTo(agentID.getName())<0))
		    return true;
	    }
	return false;
	*/
    }

    /** Get or set some information about message numbering */
    int getComputationNum() {
	return computationNum; 
    }
    void setComputationNum(int computationNum){
      this.computationNum = computationNum;
    }
    //synchronized int getMessageNum(){ return messageNum; }
    void  setMessageNum(int messageNum){
      this.messageNum = messageNum;
    }

    /**
     * Put the message on the outQueue object. This queue sends
     * its messages to the messageSender obejct, which sends
     * them to their destinations.
     *
     * @param  Message the message to send
     * @see MessageQueue
     */
    protected void putOnOutputQueue(Message message){
	// We set the ID of the message to be sent
	message.setOtherData(agentID, computationNum, messageNum);
	messageNum++;

	select.call(0, outQueue, outQueue._addMessage);
        //MySystem.out.println("GAgent:putting message on output queue:"+message);
	Synch_Wait();
        //MySystem.err.println("GAgent:put message on output queue"+message);
	outQueue.addMessage(message);
	outputQueueSize=outQueue.size();
        //MySystem.err.println("GAgent:put message on output queue:now1:"+outQueue.size());
    }

    //force if should update value-after blockings
    public int getOutputQueueSize(boolean reuse){
	if(reuse)
	    return outputQueueSize;
	return outputQueueSize=outQueue.size();
    }

    /** If there is a running protocol, kill it. */
    protected void killProtocol(){
	// Ask the protocol to stop if it is still active
	//System.out.println("Trying to shut down "+agentID+" ...");
	if(threadProtocol != null)
	    if(threadProtocol.isAlive()) {
		// Protocol is still running
		select.call(0, instanceOfUsedProtocol,
			    instanceOfUsedProtocol._terminate);
		Synch_Wait();
		instanceOfUsedProtocol.terminate();
	    }

	// The protocol message queue has no receiver
	select.call(0, protocolInQueue, protocolInQueue._setReceiver);
	Synch_Wait();
	protocolInQueue.setReceiver(null);

	instanceOfUsedProtocol = null;
	threadProtocol = null;
	System.out.println(agentID+" shut down its protocol.");
    }

    /**
     * Wait until all needed messages are send, i.e. the outQueue object
     * holds no more message, then exit.
     */
    protected void shutDown(){
	// The agent can't exit before all messages are sent
	// (Message_BrokerShutsDown, Message_Unsubscribe)
	select.call(0, outQueue, outQueue._shutDown);
	Synch_Wait();
	outQueue.shutDown();
	// Waits until all messages are sent
	do {
	    select.call(0, outQueue, outQueue._hasShutDown);
	    Synch_Wait();
	} while(!outQueue.hasShutDown());
        messageSender.end=true;
	System.out.println(agentID + " exits normally");
	System.exit(0);
    }
}
