/*
The package platform is a semester project by Galley, Michel,
that Silaghi, Marius-Calin adviced in Spring 2000.
Several modifications were done subsequently by Silaghi, Marius-Calin.
Copyright (c) 2000, 2001, 2002, 2004
*/
/**
 * @(#)Message.java  1.0 00/02/05
 *
 * Intelligent Agents Collaborating in Distributed Problem Solving
 *
 * Supervisors:   Marius-Calin Silaghi, Djamila Sam-Haroud
 * Author:        Michel Galley, Marius Silaghi
 */

package platform;

import java.lang.*;
import java.io.*;
import java.util.*;
import utils.*;

/**
 * <code>Message</code> is an abstract class representing a message and holding
 * information about its sender and its receiver, and about message numbering.
 * The messages are numbered with two fields : the computation number and the
 * message number. We can use the two fields to reorder messages, delete too
 * old messages, ...
 *
 *
 * @author Michel Galley, Marius Silaghi
 * @see Message
 * @see MessageList
 * @version 1.0 00/02/05
 */

abstract public class Message implements Serializable{
    // To identify a message, we need both senderID and messageID.
    // This pair is unique for a given computation.

    // Messages disk usage must be very small, because there can be millions
    // of messages to save on disk. For that purpose, custom data streaming
    // is used to save those messages on disk. To save even more space, agent
    // names and messages names are put into dictionaries and their key are
    // saved to disk (rather than their names).

    private static final short MESSAGE_SIZE = 14;
    private static final short MAX_MSG_LEN=2000;

    /** Dictionaries containing class names and agentNames */
    private static Vector classes;
    private static Vector agentNames;

    /** @serial Message sender's ID. */
    private AgentID senderID;
    /** @serial Message receiver's ID. */
    private AgentID receiverID;
    /** @serial Message computation number. */
    private short computationNum;
    /** @serial Message number. */
    private int messageNum;

    /** Empty constructor. */
    protected Message() {}

    /**
     * Initializes a newly created Message object. The receiver
     * is set in its constructor. Other fields are set by the setOtherData
     * function.
     *
     * @param  receiverID The receiver of this message.
     * @see Message#setOherData
     */
    protected Message(AgentID receiverID)
    {
	this.receiverID = receiverID;
    }

    /**
     * The platform must call this static function before any message is
     * sent.
     */
    static void initMessageLogging()
    {
	classes = new Vector(0);
	agentNames = new Vector(0);
    }

    /**
     * Set other important fields like the sender's ID, the computation
     * number and the message number (the message receiver is set in
     * the constructor).
     *
     * @param  senderID The ID of the sender
     * @param  computationNum The computation number
     * @param  messageNum The message number
     * @see Message
     */
    public void setOtherData(AgentID senderID, int computationNum,
			int messageNum)
    {
	this.senderID = senderID;
	this.computationNum = (short)computationNum;
	this.messageNum = messageNum;
    }

    /** Return the ID of the sender. */
    public AgentID getSenderID() { return senderID; }
    /** Return the ID of the receiver. */
    public AgentID getReceiverID() { return receiverID; }
    /** Return the computation number. */
    public int getComputationNum() { return computationNum; }
    /** Return the message number. */
    public int getMessageNum() { return messageNum; }

    /**
     * Return a string giving information about the message, i.e.
     * the ID of sender and receiver, computation number and
     * information about message's specific data (according to message's
     * class).
     */
    public String toString(){
	if(this instanceof Message_NotifyMessage) {
	    Message_NotifyMessage notify = (Message_NotifyMessage)this;

	    return new String((notify.origSenderID+"         ")
			      .toCharArray(), 0, 9)
		+ new String((notify.origReceiverID+"         ")
			     .toCharArray(), 0, 9)
		+ new String((notify.origComputationNum+"     ")
			     .toCharArray(), 0, 5)
		+ new String((notify.origMessageNum+"     ")
			     .toCharArray(), 0, 5)
		+ _toString();
	} else {
	    return new String((senderID+"         ")
			      .toCharArray(), 0, 9)
		+ new String((receiverID+"         ")
			     .toCharArray(), 0, 9)
		+ new String((computationNum+"     ").toCharArray(), 0, 5)
		+ new String((messageNum+"     ").toCharArray(), 0, 5)
		+ _toString();
	}
    }
    /**
     * All Message subclasses must implement this method which must give
     * information about their specific data.
     */
    abstract String _toString();

    /**
     * Serialize this object on a DataOutputStream object.
     *
     * @param  out The DataOutputStream to write on
     */
    void writeMessage(DataOutputStream out) throws IOException
    {
	// class information and specific member variables
	Class className = this.getClass();
	int index = classes.indexOf(className);
	if(index == -1) {
	    // Create a new entry in the dictionary
	    classes.add(className);
	    index = classes.size()-1;
	}
	// Write message's size on the stream, so the skipMessage
	// function can skip unwanted messages.
	short msgSize=(short)_size();
	ByteArrayOutputStream msgStream=
	    new ByteArrayOutputStream(msgSize);
	DataOutputStream msgDataStream = new DataOutputStream(msgStream);
	_writeMessage(msgDataStream);
	msgDataStream.flush();
	byte[] msg=msgStream.toByteArray();
	if(msgSize!=msgDataStream.size())
	    MySystem.err.println("Message declared size:"+msgSize+";and real size:"+msgDataStream.size()+" :\n"+this);
	msgSize=(short)msgDataStream.size();
	if(msgSize+ MESSAGE_SIZE-2<0){
	    MySystem.err.println("Message size reduced from "+msgSize+"meaning"+msgDataStream.size()+" to "+MAX_MSG_LEN);
	    msgSize=MAX_MSG_LEN;
	}
	out.writeShort(msgSize+ MESSAGE_SIZE-2); // 2 bytes
	out.writeShort((short)index);  // 2 bytes

	if(msg.length<msgSize){
	    MySystem.err.println("Message size is "+msgSize+" rather than "+msg.length);
	    MySystem.exit(1);
	    byte[] tmp = new byte[msgSize-msg.length];
	    out.write(msg,0,msg.length);      // msgSize bytes
	    out.write(tmp,0,tmp.length);
	}else
	    out.write(msg,0,msgSize);      // msgSize bytes

	// senderID
	index = agentNames.indexOf(senderID);
	if(index == -1) {
	    agentNames.add(senderID);
	    index = agentNames.size()-1;
	}
	out.writeShort((short)index);  // 2 bytes

	// receiverID
	index = agentNames.indexOf(receiverID);
	if(index == -1) {
	    agentNames.add(receiverID);
	    index = agentNames.size()-1;
	}
	out.writeShort((short)index);  // 2 bytes

	// other member variables
	out.writeShort(computationNum);// 2 bytes
	out.writeInt(messageNum);      // 4 bytes
    }

    /**
     * Create a new Message serialized on a DataInputStream object.
     *
     * @param  out The DataInputStream to read on
     */
    static Message readMessage(DataInputStream in) throws IOException
    {
	Message message = null;

	// class information and specific member variables
	short msgsize=in.readShort();
	short classid=in.readShort();
	Class msgClass=null;
	try{
	    msgClass = (Class)classes.get(classid);
	}catch(ArrayIndexOutOfBoundsException e){
	    MySystem.err.println("bad msgclassid index:"+classid);
	    in.skip((short)msgsize-2);
	    return null;
	}
	try { message = (Message)msgClass.newInstance(); }
	catch(Exception e) {e.printStackTrace(); System.exit(-1); }
	byte[] msg=new byte[msgsize-(MESSAGE_SIZE-2)];
	in.readFully(msg);
	ByteArrayInputStream msgStream =
	    new ByteArrayInputStream(msg);
	DataInputStream msgDataStream = new DataInputStream(msgStream);
	try{
	    message._readMessage(msgDataStream);
	}catch(Exception e){
	    e.printStackTrace();
	    MySystem.err.println("Message troubles reading:"+message+"\nfrom:"+new String(msg));
	    ByteArrayInputStream msgStream1 =
		new ByteArrayInputStream(msg);
	    DataInputStream msgDataStream1 = new DataInputStream(msgStream1);
	    short size1 = msgDataStream1.readShort();     //2
	    MySystem.err.println("First field="+size1);	    
	    byte[] str1 = new byte[size1];
	    MySystem.err.println("msg.length="+msg.length);	    
	    msgDataStream1.readFully(str1);
	    String command = new String(str1);
	    MySystem.err.println("command="+command);	    
	    size1 = msgDataStream1.readShort();           //2
	    MySystem.err.println("Second field="+size1);	    
	    str1 = new byte[size1];
	    msgDataStream1.readFully(str1);
	    String arguments = new String(str1);
	    MySystem.err.println("arguments="+arguments);	    
	}
	// other member variables
	short sendername=in.readShort();
	try{
	    message.senderID = (AgentID)agentNames.get(sendername);
	}catch(ArrayIndexOutOfBoundsException e){
	    MySystem.err.println("bad senderName index:"+sendername);
	    MySystem.err.println("at message:"+message);
	}
	short receivername=in.readShort();
	try{
	    message.receiverID = (AgentID)agentNames.get(receivername);
	}catch(ArrayIndexOutOfBoundsException e){
	    MySystem.err.println("bad receiverName index:"+receivername);
	    MySystem.err.println("at message:"+message);
	}
	message.computationNum = in.readShort();
	message.messageNum = in.readInt();

	return message;
    }

    /**
     * Skip a given number of messages on a DataInputStream object.
     *
     * @param  in The DataInputStream to read on
     * @param  nb Number of messages to skip
     */
    static void skipMessages(DataInputStream in, int nb)
	throws IOException
    {
	short size;

	for(int i=0; i<nb; i++) {
 	    size = in.readShort();
	    if(size<0){
		MySystem.err.println("Trying to skip:"+size);
		MySystem.exit(1);
	    }
 	    in.skip((short)size);
	}
    }

    /**
     * Each subclass must implement this method which writes subclasses
     * specific member variables on a DataOutputStream.
     *
     * @param  out The DataOutputStream to write on
     */
    abstract void _writeMessage(DataOutputStream out) throws IOException;
    /**
     * Each subclass must implement this method which gives the size
     * in bytes of the data written on a DataOutputStream when calling
     * _writeMessage.
     */
    abstract int _size();
    /**
     * Each subclass must implement this method which reads subclasses
     * specific member variables from a DataInputStream.
     *
     * @param in The DataInputStream to read on
     */
    abstract void _readMessage(DataInputStream in) throws IOException;
}

