/*
The package platform is a semester project by Galley, Michel,
that Silaghi, Marius-Calin adviced in Spring 2000.
Several modifications were done subsequently by Silaghi, Marius-Calin.
Copyright (c) 2000, 2001, 2002, 2004
*/
/*
 * @(#)MessageList.java  1.0 00/02/04
 *
 * Intelligent Agents Collaborating in Distributed Problem Solving
 *
 * Supervisors:   Marius-Calin Silaghi, Djamila Sam-Haroud
 * Author:        Michel Galley
 */

package platform;

import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;


/**
 * This class stores all the messages sent by or sent to the broker.
 * It implements ListModel, so it can be used by the broker GUI's JList
 * component as a model. Only a part of the messages are directly available
 * (a part of the buffer). To access other messages, use the moveSelection
 * method.
 *
 * <p>This class is only meant to be accessed by the broker's GUI. It's
 * methods reflects the GUI. There is one panel showing only a part of the
 * messages and two buttons enabling the user to go forth or back.
 *
 * <p>Only a part of all messages are kept in memory, because there can be
 * millions of messages, so it is advisable to store them on disk. If you
 * intend to use moveSelection widely, keep in mind that it can cause
 * some delay.
 *
 * @author Michel Galley
 * @see ChildMessageList
 * @see BrokerGUI
 * @version 1.0 00/02/04
 */

class MessageList implements ListModel{
    /** Name of the file in which the messages are stored */
    private static final String FILE_NAME = ".history.tmp";
    static final String EMPTY_LINE = " ";
    static final int BUFFER_SIZE = 300;
    static final int VISIBLE_SIZE = 15;
    private boolean noListClear = true;
    /**
     * A buffer containing a portion of the messages stored
     * in the file.
     */
    Message[] messageBuffer;
    /** Number of messages in the buffer  */
    int nbOfMessages;
    /** Buffer's absolute start position */
    int bufferStart;
    /** Buffer's absolute end position */
    int bufferEnd;
    /** Selected message position (relative to bufferStart) */
    int selectedMessage;
    private Hashtable childMessageLists = new Hashtable();

    LogFile logFile = new LogFile();

    /** The listeners : JList components */
    private Vector listeners;

    /**
     * Initializes a newly created MessageList object
     * containing no message.
     */
    MessageList(){
	messageBuffer = new Message[BUFFER_SIZE];
	listeners = new Vector(0);
	nbOfMessages = bufferStart = selectedMessage = 0;
	bufferEnd = -1;
    }

    /** Get the ChildMessageList corresponding to the given agent */
    ChildMessageList getChildMessageList(AgentID agentID){
        return (ChildMessageList)childMessageLists.get(agentID.getName());
    }

    /**
     * Add a message to the list to the list. It is added to the file
     * and to the buffer if needed.
     *
     * @param  message The message added.
     */
    synchronized void addMessage(Message message){
	int newIndex;

	// If the message is a setColor or a initSetColor message,
	// give it to the appropriate ChildMessageList
	// (if the corresponding ChildMessageList object isn't already
	// created, create it)

	if(message instanceof Message_Subscription) {
	    if((!noListClear)||
	       (childMessageLists.get(message.getSenderID().getName())==null))
		childMessageLists.put(message.getSenderID().getName(),
				      new ChildMessageList(this,
						     message.getSenderID()));
	}
	else if(message instanceof Message_CreateColorMatrix ||
		message instanceof Message_SetColor) {
	    ChildMessageList childModel = (ChildMessageList)
		childMessageLists.get(message.getSenderID().getName());
	    childModel.processMessage(message);
	}

	// Add the message to the file
	logFile.putMessage(message);
	nbOfMessages++;

	// Put the message in the buffer if needed
	if(bufferEnd-bufferStart+1 < BUFFER_SIZE) {
	    messageBuffer[++bufferEnd-bufferStart] = message;
	    notifyBufferModification();
	}

	// Notify the listeners (view)
	for(int i=0; i<listeners.size(); i++) {
	    ListDataListener listener
		= (ListDataListener)listeners.elementAt(i);
	    newIndex = nbOfMessages - bufferStart
		- selectedMessage + VISIBLE_SIZE/2;
	    if(newIndex <= VISIBLE_SIZE)
		listener.intervalAdded
		    (new ListDataEvent(this, ListDataEvent.INTERVAL_ADDED,
				       newIndex-1, newIndex-1));
	}
	//show movie;
	try{
	    moveSelection(1);
	}catch(RuntimeException e){
	    System.err.println(message);
	    e.printStackTrace();
	    //System.exit(1);
	    throw e;
	}
    }

    /** Clears the list. */
    synchronized void clear(){
	//never clear the list
	if(noListClear) return;
	int oldSize = nbOfMessages;

	logFile.clear();
	nbOfMessages = bufferStart = selectedMessage = 0;
	bufferEnd = -1;

	notifyClear();

	if(oldSize > 0)
	    for(int i=0; i<listeners.size(); i++) {
		ListDataListener listener
		    = (ListDataListener)listeners.elementAt(i);
		listener.contentsChanged
		    (new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED,
				       0, VISIBLE_SIZE-1));
	    }
    }

    /**
     * Move the message selection in the buffer.
     *
     * @param  offset Number of messages to skip (signed value)
     */
    public boolean moveSelection(int offset){
	if(selectedMessage + offset < 0 ||
	   bufferStart + selectedMessage + offset >= nbOfMessages)
	    return false;
	selectedMessage += offset;

	// Refill the buffer if needed
	if(selectedMessage - VISIBLE_SIZE/2 < 0)
	    if(bufferStart > 0)
		fillBuffer(bufferStart + selectedMessage - BUFFER_SIZE/2);

	if(selectedMessage + VISIBLE_SIZE/2 > bufferEnd - bufferStart + 1)
	    if(bufferEnd < nbOfMessages-1)
		fillBuffer(bufferStart + selectedMessage - BUFFER_SIZE/2);

	// Update views
	for(int i=0; i<listeners.size(); i++) {
	    ListDataListener listener
		= (ListDataListener)listeners.elementAt(i);
	    listener.contentsChanged
		(new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED,
				   0, VISIBLE_SIZE-1));
	}

	// Update all children views
	Iterator keys = (Iterator)childMessageLists.keys();
	while(keys.hasNext())
	    ((ChildMessageList)childMessageLists
	     .get(keys.next())).updateListView();

	return true;
    }

    /**
     * Fill the buffer from file.
     *
     * @param  offset File position f
     */
    void fillBuffer(int offset){
	// Determine start and end position
	int oldBufferStart = bufferStart;
	bufferStart = offset;
	if(bufferStart < 0) bufferStart = 0;
	bufferEnd = bufferStart + BUFFER_SIZE-1;
	if(bufferEnd >= nbOfMessages) bufferEnd = nbOfMessages-1;

	// Fill the buffer
	bufferEnd = logFile.fillBuffer(messageBuffer, bufferStart)
	    + bufferStart;

	boolean changed=(oldBufferStart != bufferStart);
	// Find the new selected message position
	selectedMessage += (oldBufferStart - bufferStart);
	if (changed) {
	    Iterator keys = (Iterator) childMessageLists.keys();
	    while(keys.hasNext()) {
		((ChildMessageList)childMessageLists
		 .get(keys.next())).cancelSelection();
	    }
	    notifyBufferModification();
	}
    }

    /**
     * Notify all ChildMessageList objects that the
     * buffer has been modified.
     */
    void notifyBufferModification(){
	Iterator keys = (Iterator) childMessageLists.keys();
	while(keys.hasNext()) {
	    ((ChildMessageList)childMessageLists
	     .get(keys.next())).updateFromBuffer();
	}
    }

    /**
     * Notify all ChildMessageList objects that the
     * buffer has cleared.
     */
    void notifyClear(){
	Iterator keys = (Iterator) childMessageLists.keys();
	while(keys.hasNext()) {
	    ((ChildMessageList)childMessageLists
	     .get(keys.next())).clear();
	}
    }

   /**************************************************************
     * ListModel implementation :
     * Broker GUI methods : a JList listens to the history changes:
     **************************************************************/

    /**
     * Add a listener to the list that's notified each time a change to the
     * data model occurs. In the current implementation, there can be only
     * one listener.
     *
     * @param  listener The listener to add
     * @see ListData
     */
    public void addListDataListener(ListDataListener listener)
    { listeners.add(listener); }
    /**
     * Remove the listener from the list.
     *
     * @param  listener The listener to remove
     * @see ListData
     */
    public void removeListDataListener(ListDataListener listener)
    { listeners.remove(listener); }
    /**
     * Get the size of the visible part of the message buffer. It is
     * a constant value.
     */
    public int getSize() { return VISIBLE_SIZE; }
    /**
     * Get an element from the visible part of the buffer.
     *
     * @param  index Index of the element to access
     */
    public Object getElementAt(int index){
	int indexTable = selectedMessage-VISIBLE_SIZE/2+index;
	if(indexTable < 0 || indexTable > bufferEnd - bufferStart)
	    return EMPTY_LINE;
	return messageBuffer[indexTable];
    }
}

