/*
The package platform is a semester project by Galley, Michel,
that Silaghi, Marius-Calin adviced in Spring 2000.
Several modifications were done subsequently by Silaghi, Marius-Calin.
Copyright (c) 2000, 2001, 2002, 2004
*/
/*
 * @(#)AgentPlatform.java  1.0 00/02/04
 *
 * Intelligent Agents Collaborating in Distributed Problem Solving
 *
 * Supervisors:   Marius-Calin Silaghi, Djamila Sam-Haroud
 * Author:        Michel Galley
 */

package platform;

import csp.*;
import sjavap.*;
import utils.*;
import java.io.*;
import java.net.*;
import java.awt.*;
import java.awt.geom.*;

/**
 * Non-active objects and objects which don't use the sJava library can't
 * directly call sJava sychronized methods. It is however possible to bypass
 * this problem with the use of an interface. This class defines such an
 * interface.
 * This class holds all the methods enabling non-active objects to
 * call methods on a Broker or an Agent object.
 * Important notice: Platform's methods are locking as long as
 * the called object is not ready.
 *
 * @author Michel Galley
 * @see GenericAgent
 * @see Broker
 * @see Agent
 * @version 1.0 00/02/04
 */

public class Platform implements Reply{
    private GenericAgent genericAgent = null;
    private Broker broker = null;
    private Agent agent = null;
    private boolean inWait = false;

    /**    active object support (needs implements Reply)               **/
    /*********************************************************************/
    /* Written by Claude Petitpierre,   LTI-DI-EPFL,           7 Apr 1999*/
    /*********************************************************************/
    /* Modif. by Marius-Calin Silaghi, LIA-DI-EPFL,silaghi@lia.di.epfl.ch*/
    /*  Bug that may lead to deadlock in Synch_Wait(),fixed,  30 Nov 1999*/
    /*********************************************************************/
    /**/ Synch_List select = new Synch_List(this) ;                      //
    /**/ synchronized public void Synch_Reply()                          //
    /**/   { notifyAll(); }                                                 //
    /**/ int now() {return Kernel.kern.currentTime;}                     //
    /**/ int Synch_Wait() {                                              //
    /**/   Kernel.kern.Synch_Request(select);                            //
    /**/   do                                                            //
    /**/     {try {wait();} catch(InterruptedException e){}}             //
    /**/   while(select.nbEntries!=0);                                   //
    /**/   return select.Value; }                                        //
    /*********************** Need not be modified ************************/

    /**
     * Platform for the broker. It initializes a newly created Platform object.
     * It creates a new Broker object, on which the methods will be called.
     *
     * @param  knownProtocols Array of the classes of the protocols known by
     * the broker.
     * @param  withGUI Indicates if the broker must create a GUI
     */
    public Platform(Class[] knownProtocols, boolean withGUI){
	// Broker constructor:
	broker = new Broker(knownProtocols, withGUI);
	this.broker = broker;
	genericAgent = broker;
	new Thread(broker).start();
    }

    /**
     * Platform for the broker. Initializes a newly created Platform object.
     * The Broker object passed as a parameter will be the receiver of the
     * calls.
     *
     * @param broker Broker on which the methods will be called
     */
    public Platform(Broker broker) { this.broker = broker; }

    /**
     * Platform for the agents. Initializes a newly created Platform object.
     * It creates a new Agent object, on which the methods will be called.
     *
     * @param  agentName The name of the agent
     * @param  knownProtocols Array of the classes of the protocols known by
     * the agent.
     * @param  broker_host IP address of the broker
     * @param  broker_port Port number of the broker
     * @param  withGUI Indicates if the agent must create a GUI
     */
    public Platform(String agentName,
		    Class[] knownProtocols,
		    InetAddress broker_host, int broker_port,
		    boolean withGUI){
	// Agent constructor:
	agent = new Agent(agentName, knownProtocols,
			  broker_host, broker_port, withGUI);
	this.agent = agent;
	genericAgent = agent;
	new Thread(agent).start();
    }

    /**
     * Platform for the agents. Initializes a newly created Platform object.
     * The Agent object passed as a parameter will be the receiver of the
     * calls.
     *
     * @param agent Agent on which the methods will be called
     */
    public Platform(Agent agent) { this.agent = agent; }


    /*********************
     * GENERIC AGENT CALLS
     *********************/

    /**
     * Return the ID of the agent (AgentID object).
     */
    public AgentID getID() { return genericAgent.getID(); }

    /**
     * The broker sends a message to the agent whose ID
     * is receiverID. The message format is :
     * <command> [<arg1> ... <argn>]
     *
     * @param  command Command serializable object
     * @param  arguments Arguments serializable object
     * @param  receiverID Destination of the object
     */
    synchronized public void sendMessage(Serializable command,
					 Serializable arguments,
					 AgentID receiverID){
	if(broker==null)
	    return;
	// Wait until the inWait lock is free
	try {
	    while(inWait){
		wait();
		//if(inWait)
		//    notify();
	    }
	} catch(Exception e) {
	    e.printStackTrace();
	}
	inWait = true;
	select.call(0, broker, broker._sendMessage);
	Synch_Wait();
	inWait = false;
        //MySystem.err.println("Platform:ask broker load for"+receiverID);
	broker.sendMessage(command, arguments, receiverID);
	notify();
    }

    /**
     * The genericAgent sends a message to the genericAgent whose ID
     * is receiverID. The message format is :
     * <command> [<arg1> ... <argn>]
     *
     * @param  command Command serializable object
     * @param  arguments Arguments serializable object
     * @param  receiverID Destination of the object
     */
    synchronized public void sendCommand(Serializable command,
					 Serializable arguments,
					 AgentID receiverID)
    {
	// Wait until the inWait lock is free
	try { while(inWait){ wait();notify();} } catch(Exception e) {}
	inWait = true;
	select.call(0, genericAgent, genericAgent._sendCommand);
	Synch_Wait();
	inWait = false;
	genericAgent.sendCommand(command, arguments, receiverID);
    }

    /**
     * Set the problem the genericAgent has to solve.
     *
     * @param  file The file containing the problem
     */
    synchronized public boolean setProblem(File file)
    {
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, genericAgent, genericAgent._setProblem);
	Synch_Wait();
	inWait = false;
	return genericAgent.setProblem(file);
    }

    /**
     * Return The problem the genericAgent has to solve
     */
    synchronized public Problem getProblem()
    {
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, genericAgent, genericAgent._getProblem);
	Synch_Wait();
	inWait = false;
	return genericAgent.getProblem();
    }

    /**************
     * BROKER CALLS
     **************/

    /**
     * Return an array containing all agent IDs.
     */
    synchronized public AgentID[] getAgentIDs(){
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, broker, broker._getAgentIDs);
	Synch_Wait();
	inWait = false;
	return broker.getAgentIDs();
    }
    synchronized public void removeSubscriptions(){
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, broker, broker._removeSubscriptions);
	Synch_Wait();
	inWait = false;
	broker.removeSubscriptions();
    }

    /**
     * Sets the maxium number of subscribed agents. If this function is
     * not called, we assume that this number is Integer.MAX_VALUE.
     * If a protocol has been chosen, this protocol is started as soon
     * as the maximum number of subscribed agents is reached.
     * This method cannot be called during a computation.
     *
     * @param  nbOfAgents Maximum number of agents
     */
    synchronized public void setMaxNbOfAgents(int nbOfAgents)
    {
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, broker, broker._setMaxNbOfAgents);
	Synch_Wait();
	inWait = false;
	broker.setMaxNbOfAgents(nbOfAgents);
    }

    /**
     * Return an array containing the protocols known by all participants.
     */
    synchronized public Class[] getProtocols()
    {
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, broker, broker._getProtocols);
	Synch_Wait();
	inWait = false;
	return broker.getProtocols();
    }

    /**
     * Let the caller choose a protocol among the protocols
     * returned by getProtocols. If the protocol is not known by all
     * participants, chooseProtocol returns false. Otherwise, it
     * rerturns true.
     *
     * @param  protocol The chosen protocol
     */
    synchronized public boolean chooseProtocol(Class protocol)
    {
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, broker, broker._chooseProtocol);
	Synch_Wait();
	inWait = false;
	return broker.chooseProtocol(protocol);
    }

    /**
     * The protocol is started. The broker sends a message to all subscribers
     * asking them to start their protocols (the same class).
     */
    synchronized public void startProtocol(BrokerOwner owner){
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, broker, broker._startProtocol);
	Synch_Wait();
	inWait = false;
	broker.startProtocol(owner);
    }

    /**
     * The protocol is stopped. The broker sends a message to all subscribers
     * asking them to stop their protocols.
     */
    synchronized public void endProtocol()
    {
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, broker, broker._endProtocol);
	Synch_Wait();
	inWait = false;
	broker.endProtocol();
    }

    /**
     * The broker shuts down. The broker sends a message to all
     * subscribers telling them that the broker is going to shut down.
     */
    synchronized public void brokerShutsDown()
    {
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, broker, broker._brokerShutsDown);
	Synch_Wait();
	inWait = false;
	broker.brokerShutsDown();
    }

    /**
     * Only protocols can call this method. It informs the broker that
     * the protocol has ended.
     */
    synchronized public void protocolFinished(Object result){
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, broker, broker._protocolFinished);
	Synch_Wait();
	inWait = false;
	broker.protocolFinished(result);
    }

    /*************
     * AGENT CALLS
     *************/

    /**
     * The ID of the broker.
     */
    public AgentID getBrokerID() { return agent.getBrokerID(); }

    /**
     * Register the agent to the broker. The agent will be considered
     * as a participant in the computations. The agent submits all the
     * protocols it knows (the protocols are passed to the constructor).
     */
    synchronized public void subscribe()
    {
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, agent, agent._subscribe);
	Synch_Wait();
	inWait = false;
	agent.subscribe();
    }

    /**
     * Unregister the agent to the broker. The agent won't receive
     * messages anymore.
     */
    synchronized public void unsubscribe()
    {
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, agent, agent._unsubscribe);
	Synch_Wait();
	inWait = false;
	agent.unsubscribe();
    }

    /**
     * Send a request to the broker to create a matrix representing the
     * constraint shared by two variables (this matrix is shown in the GUI).
     * Call on this method is a prerequisite to any call of a setColor
     * involving this matrix.
     *
     * @param  var1 Name of the first variable
     * @param  var2 Name of the second variable
     * @param  sizeX Width of the matrix
     * @param  sizeY Height of the matrix
     * @see Platform#setColor
     */
    synchronized public void createColorMatrix(String var1, String var2,
					       int sizeX, int sizeY)
    {
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, agent, agent._createColorMatrix);
	Synch_Wait();
	inWait = false;
	agent.createColorMatrix(var1, var2, sizeX, sizeY);
    }

    /**
     * Put a new color in the cell positioned at (posX, posY) in the
     * matrix (of the GUI) involving both var1 and var2 variables.
     *
     * @param  var1 Name of the first variable
     * @param  var2 Name of the second variable
     * @param  sizeX Width of the matrix
     * @param  sizeY Height of the matrix
     * @param  color Cell's new color
     * @param  oldColor Cell's old color
     * @see Platform#createColorMatrix
     */
    synchronized public void setColor(String var1, String var2,
				      int posX, int posY,
				      Color color, Color oldColor)
    {
	try { while(inWait) wait(); } catch(Exception e) {}
	inWait = true;
	select.call(0, agent, agent._setColor);
	Synch_Wait();
	inWait = false;
	agent.setColor(var1, var2, posX, posY, color, oldColor);
    }
}
