/*
The package platform is a semester project by Galley, Michel,
that Silaghi, Marius-Calin adviced in Spring 2000.
Several modifications were done subsequently by Silaghi, Marius-Calin.
Copyright (c) 2000, 2001, 2002, 2004
*/
/**
 * @(#)Protocol.java  1.0 00/02/06
 *
 * Intelligent Agents Collaborating in Distributed Problem Solving
 *
 * Supervisors:   Marius-Calin Silaghi, Djamila Sam-Haroud
 * Author:        Michel Galley
 */

package platform;

import csp.*;
import utils.*;
import sjavap.*;
import java.io.*;

/**
 * The <code>Protocol</code> abstract class provides the protocol
 * developper with some basic methods helping him to build protocols.
 * It also defines abstract methods the programmer must implement in
 * order to establish communication means between the agent and it's
 * protocol.
 * <p>Please refer to the "Intelligent Agents Collaborating in Distributed
 * Problem Solving" paper for further information. An entire chapter covers
 * the writing of protocols.
 *
 * @author Michel Galley
 * @see Platform
 * @see MessageReceiver
 * @version 1.0 00/02/06
 */

public abstract class Protocol implements Runnable, MessageReceiver{
    /** sJava identification for the getMessage method. */
    //public static final int _getMessage = 1;
    /** sJava identification for the getMessage method. */
    public static final int _terminate = 2;

    /** The GenricAgent that created this protocol. */
    protected GenericAgent genericAgent;
    /**
     * All agent ID's. Only the protocol running for an agent has
     * information about it.
     */
    protected AgentID[] agentIDs;
    /**
     * Queue containing protocol-dependent messages. The protocol's
     * getMessage is triggered off each time a new message appears
     * on the queue.
     */
    protected MessageQueue queue;
    /** The problem the agent has to solve. */
    protected Problem problem;

    /**    active object support (needs implements Reply)               **/
    /*********************************************************************/
    /* Written by Claude Petitpierre,   LTI-DI-EPFL,           7 Apr 1999*/
    /*********************************************************************/
    /* Modif. by Marius-Calin Silaghi, LIA-DI-EPFL,silaghi@lia.di.epfl.ch*/
    /*  Bug that may lead to deadlock in Synch_Wait(),fixed,  30 Nov 1999*/
    /*********************************************************************/
    /**/ public Synch_List select = new Synch_List(this) ;               //
    /**/ synchronized public void Synch_Reply()                          //
    /**/   { notify(); }                                                 //
    /**/ public int now() {return Kernel.kern.currentTime;}              //
    /**/ public int Synch_Wait() {                                       //
    /**/   Kernel.kern.Synch_Request(select);                            //
    /**/   do                                                            //
    /**/     {try {wait();} catch(InterruptedException e){}}             //
    /**/   while(select.nbEntries!=0);                                   //
    /**/   return select.Value; }                                        //
    /*********************** Need not be modified ************************/

    public Object startParameter;

    /**
     * Initializes a newly created Protocol object. All member variables
     * are set to null.
     */
    public Protocol() {}
    /**
     * Initializes a newly created broker Protocol object. Only a broker
     * can call this constructor. It must provide information about the
     * queue containing protocol-dependent messages and an array containing
     * all subscribed agent IDs.
     *
     * @param  broker The broker calling this constructor
     * @param  problem The broker may provide information about a problem
     * (optional parameter, it can be null)
     * @param  queue Queue containing protocol-dependent messages
     * @param  agents An array of all agent ID's
     */
    Protocol(Broker broker, Problem problem, MessageQueue queue,
	     AgentID[] agents)
    { setParameters(broker, problem, queue, agents); }
    /**
     * Initializes a newly created agent Protocol object. Only a agnet
     * can call this constructor. It must provide information about the
     * queue containing protocol-dependent messages.
     *
     * @param  agent The agent calling this constructor
     * @param  problem The problem the agent has to solve
     * @param  queue Queue containing protocol-dependent messages
     */
    Protocol(Agent agent, Problem problem, MessageQueue queue)
    { setParameters(agent, problem, queue, null); }

    /****************
     * Public methods
     ****************/

    protected Class getAgentProtocolClass(){
	return getClass();
    }
    protected Class getBrokerProtocolClass(){
	return getClass();
    }

    /**
     * Provides a way to set all needed instance variables.
     *
     * @param  genericAgent The genericAgent calling this method
     * @param  problem The problem to solve
     * @param  queue Queue containing protocol-dependent messages
     * @param  agents An array of all agent ID's
     */
    void setParameters(GenericAgent genericAgent, Problem problem,
		       MessageQueue queue, AgentID[] agentIDs){
	this.genericAgent = genericAgent;
	this.problem = problem;
	this.queue = queue;
	this.agentIDs = agentIDs;
    }

    /**
     * Two protocols are equals if they are instances of the same class.
     *
     * @param  object The object compared to this protocol.
     */
    public boolean equals(Object object)
    {
	if(object instanceof Protocol)
	    return (this.getClass() == object.getClass());
	return super.equals(object);
    }


    /**
     * The server gives all protocol-dependent messages it gets to the
     * a this protocol (via a message queue). This method must be defined
     * as synchronized. This method is called as soon as a new message is
     * available. The role of this method is to process the new message. But
     * beware, do not call Synch_Wait() methods in getMessage.
     *
     * @param  message The new message
     */
    abstract public void getMessage(Message message);

    /**
     * Ask the protocol to stop. The implementation of this method must
     * ensure that the protocol is going to stop in a very near future.
     * This method must be defined as synchronized. It is advisable not
     * to enter a waiting state without accepting a call to this method.
     * If this condition isn't satisfied, the agent might be unable to
     * go on and participate to a new computation the broker has started.
     */
    abstract public void terminate();
    /** Return the protocol's name. */
    abstract public String toString();

    /*******************
     * Protected methods
     *******************/

    /**
     * Initializes the protocol. In the current implementation, the protocol
     * simply subscribs to the message queue as a MessageReceiver object.
     * Any protocol should call this method when it starts.
     */
    protected void initProtocol()
    {
	select.call(0, queue, queue._setReceiver);
	Synch_Wait();
	queue.setReceiver(this);
    }

    /**
     * Cleans up the protocol. In the current implementation, the protocol
     * simply notifies the agent of the protocol stopping. Any protocol should
     * call this method before it stops.
     */
    protected void cleanUpProtocol(Object result){
	if(genericAgent instanceof Broker) {
	    Broker broker = (Broker)genericAgent;
	    select.call(0, broker, broker._protocolFinished);
	    Synch_Wait();
	    broker.protocolFinished(result);
	}
    }
    protected void cleanUpProtocol(){
      cleanUpProtocol(null);
    }
}
